"""Composable FastAPI mixins for LlamaBot."""

from fastapi import HTTPException


class APIMixin:
    """This is a FastAPI mixin for LlamaBot.

    By inheriting from this class, you can add FastAPI endpoints to your bot.
    The default API endpoint that we add is the /generate endpoint,
    which accepts the a string and returns a string.
    """

    def create_endpoint(self):
        """Create an endpoint for the bot."""

        async def endpoint(request: str):
            """Endpoint for the bot."""
            try:
                response = self(request)
                return {"response": response.content}
            except Exception as e:
                raise HTTPException(status_code=500, detail=str(e))

        return endpoint



"""Llamabot Zotero CLI."""

from datetime import date
from pathlib import Path

import typer
from dotenv import load_dotenv
from slugify import slugify

from llamabot import QueryBot
from llamabot.recorder import PromptRecorder
from llamabot.zotero.library import ZoteroItem, ZoteroLibrary
from llamabot.zotero.completer import PaperTitleCompleter
from llamabot.prompt_library.zotero import paper_summary, docbot_sysprompt
from llamabot.config import default_language_model
from .utils import configure_environment_variable, exit_if_asked, uniform_prompt

load_dotenv()

app = typer.Typer()

ZOTERO_JSON_DIR = Path.home() / ".llamabot/zotero/zotero_index/"
ZOTERO_JSON_DIR.mkdir(parents=True, exist_ok=True)

try:
    from rich.console import Console
    from rich.progress import Progress, SpinnerColumn, TextColumn
except ImportError:
    raise ImportError(
        "rich is not installed. Please install it with `pip install llamabot[cli]`."
    )

progress = Progress(
    SpinnerColumn(),
    TextColumn("[progress.description]{task.description}"),
    transient=False,
)

console = Console()


@app.command()
def configure(
    library_id: str = typer.Option(..., prompt=True),
    api_key: str = typer.Option(..., prompt=True),
    library_type: str = typer.Option(default="user", prompt=True),
):
    """Configure Llamabot Zotero CLI environment variables.

    :param library_id: Zotero library ID
    :param api_key: Zotero API key
    :param library_type: Zotero library type
    """
    configure_environment_variable("ZOTERO_LIBRARY_ID", library_id)
    configure_environment_variable("ZOTERO_API_KEY", api_key)
    configure_environment_variable("ZOTERO_LIBRARY_TYPE", library_type)


@app.command()
def chat(
    query: str = typer.Argument("", help="The thing you want to chat about."),
    sync: bool = typer.Option(
        True, help="Whether or not to synchronize the Zotero library."
    ),
    model_name: str = default_language_model(),
):
    """Chat with a paper.

    :param query: A paper to search for, whether by title, author, or other metadata.
    :param sync: Whether or not to synchronize the Zotero library.
    :param model_name: The name of the model to use.
    """
    try:
        from caseconverter import snakecase
    except ImportError:
        raise ImportError(
            "caseconverter is not installed. Please install it with `pip install llamabot[cli]`."
        )

    try:
        from prompt_toolkit import prompt
    except ImportError:
        raise ImportError(
            "prompt_toolkit is not installed. Please install it with `pip install llamabot[cli]`."
        )

    typer.echo("Llamabot Zotero Chatbot initializing...")
    typer.echo("Use Ctrl+C to exit anytime.")

    if sync:
        library = ZoteroLibrary()
        library.to_json(ZOTERO_JSON_DIR)
    else:
        library = ZoteroLibrary(json_dir=ZOTERO_JSON_DIR)

    completer = PaperTitleCompleter(library.key_title_map().values())
    while True:
        user_choice = prompt(
            "Please choose an paper: ",
            completer=completer,
            complete_while_typing=True,
        )
        if user_choice in library.key_title_map().values():
            break
    typer.echo(f"Awesome! You have chosen the paper: {user_choice}")

    paper_key = library.key_title_map(inverse=True)[user_choice.strip(" ")]

    # Retrieve paper from library
    with progress:
        task = progress.add_task("Downloading paper...")
        entry: ZoteroItem = library[paper_key]
        fpath = entry.download_pdf(Path("/tmp"))
        progress.remove_task(task)

    typer.echo(f"Downloaded paper to {fpath}")

    with progress:
        task = progress.add_task("Embedding paper and initializing bot...")
        docbot = QueryBot(
            docbot_sysprompt(),
            collection_name=slugify(user_choice)[:63],
            document_paths=[fpath],
            model_name=model_name,
        )
        progress.remove_task(task)

    # From this point onwards, we need to record the chat.
    pr = PromptRecorder()
    date_str = date.today().strftime("%Y%m%d")
    snaked_user_choice = f"{snakecase(user_choice)}"
    save_path = Path(f"{date_str}_{snaked_user_choice}.md")
    with pr:
        typer.echo("\n\n")
        typer.echo("Here is a summary of the paper for you to get going:")
        docbot(paper_summary())
        typer.echo("\n\n")
        pr.save(save_path)

    while True:
        with pr:
            query = uniform_prompt()
            exit_if_asked(query)
            docbot(query)
            typer.echo("\n\n")

            # Want to append YYYYMMDD before filename.
            pr.save(save_path)



"""Class definition for QueryBot."""

import contextvars
from pathlib import Path
from typing import Optional, Union
from dotenv import load_dotenv

from llamabot.config import default_language_model

from llamabot.bot.simplebot import SimpleBot
from llamabot.components.messages import AIMessage, BaseMessage, HumanMessage
from llamabot.components.docstore import LanceDBDocStore
from llamabot.components.chatui import ChatUIMixin
from llamabot.components.messages import (
    RetrievedMessage,
)
from slugify import slugify

load_dotenv()


CACHE_DIR = Path.home() / ".llamabot" / "cache"
prompt_recorder_var = contextvars.ContextVar("prompt_recorder")


class QueryBot(SimpleBot, ChatUIMixin):
    """Initialize QueryBot.

    :param system_prompt: The system prompt to use.
    :param collection_name: The name of the collection to use.
    :param document_paths: The paths to the documents to use.
    :param docstore_type: The type of document store to use ("lancedb", "sqlitevec", etc.)
    :param mock_response: A mock response to use for testing.
    :param stream_target: The target to stream to ("panel" or "stdout").
    """

    def __init__(
        self,
        system_prompt: str,
        collection_name: str,
        initial_message: Optional[str] = None,
        document_paths: Optional[Path | list[Path]] = None,
        docstore_type: str = "lancedb",  # Add this parameter
        mock_response: str | None = None,
        temperature: float = 0.0,
        model_name: str = default_language_model(),
        stream_target: str = "stdout",
        **kwargs,
    ):
        SimpleBot.__init__(
            self,
            system_prompt=system_prompt,
            temperature=temperature,
            model_name=model_name,
            mock_response=mock_response,
            stream_target=stream_target,
            **kwargs,
        )

        collection_name = slugify(collection_name, separator="_")

        # Initialize the appropriate document store
        if docstore_type == "lancedb":
            self.docstore = LanceDBDocStore(
                table_name=collection_name,
                storage_path=Path.home() / ".llamabot" / "lancedb",
            )
        else:
            raise ValueError(f"Unknown docstore type: {docstore_type}")

        # Add documents to the store
        if document_paths:
            self.docstore.add_documents(document_paths=document_paths)

        self.response_budget = 2_000

        ChatUIMixin.__init__(self, initial_message)

    def __call__(
        self, query: Union[str, HumanMessage, BaseMessage], n_results: int = 20
    ) -> AIMessage:
        """Query documents within QueryBot's document store.

        We use RAG to query out documents.

        :param query: The query to make of the documents.
        """
        messages = [self.system_prompt]

        retreived_messages = set()

        q = query
        if isinstance(query, (BaseMessage, HumanMessage)):
            q = query.content
        retrieved_messages = retreived_messages.union(
            self.docstore.retrieve(q, n_results)
        )
        retrieved = [RetrievedMessage(content=chunk) for chunk in retrieved_messages]
        messages.extend(retrieved)
        messages.append(HumanMessage(content=q))
        if self.stream_target == "stdout":
            response: AIMessage = self.stream_stdout(messages)
            return response
        elif self.stream_target == "panel":
            return self.stream_panel(messages)



"""
This module provides a set of tests for the PromptRecorder
and autorecord functions in the llamabot.recorder module.
"""

import pandas as pd
import pytest
from sqlalchemy import create_engine, inspect
from sqlalchemy.orm import sessionmaker
from llamabot.recorder import (
    PromptRecorder,
    autorecord,
    MessageLog,
    Base,
    upgrade_database,
    add_column,
    sqlite_log,
)
from llamabot.components.messages import BaseMessage
from unittest.mock import Mock


@pytest.fixture
def engine():
    """Create a temporary in-memory SQLite database for testing."""
    return create_engine("sqlite:///:memory:")


@pytest.fixture
def session(engine):
    """Create a new session for testing."""
    Base.metadata.create_all(engine)
    Session = sessionmaker(bind=engine)
    return Session()


def test_init():
    """Test that the PromptRecorder is initialized correctly."""
    recorder = PromptRecorder()
    assert recorder.prompts == []


def test_log():
    """Test that the PromptRecorder logs prompts and responses correctly."""
    recorder = PromptRecorder()
    recorder.log("prompt1", "response1")
    assert recorder.prompts == [{"prompt": "prompt1", "response": "response1"}]


def test_repr():
    """Test that the PromptRecorder is represented correctly."""
    recorder = PromptRecorder()
    recorder.log("prompt1", "response1")
    assert isinstance(recorder.__repr__(), str)


def test_dataframe():
    """Test that the PromptRecorder is represented correctly as a dataframe."""
    recorder = PromptRecorder()
    recorder.log("prompt1", "response1")
    assert isinstance(recorder.dataframe(), pd.DataFrame)


def test_autorecord_no_context():
    """Test that autorecord works without a context."""
    prompt = "Hello"
    response = "Hi"
    autorecord(prompt, response)


def test_autorecord_with_context():
    """Test that autorecord works with a context."""
    prompt = "Hello"
    response = "Hi"
    recorder = PromptRecorder()
    with recorder:
        autorecord(prompt, response)
    assert len(recorder.prompts) == 1
    assert recorder.prompts[0]["prompt"] == prompt
    assert recorder.prompts[0]["response"] == response


def test_autorecord_multiple_with_context():
    """Test that autorecord works with a context multiple times."""
    prompts = ["Hello", "How are you?"]
    responses = ["Hi", "I'm good, thanks!"]
    recorder = PromptRecorder()
    with recorder:
        for p, r in zip(prompts, responses):
            autorecord(p, r)
    assert len(recorder.prompts) == 2
    for i, (p, r) in enumerate(zip(prompts, responses)):
        assert recorder.prompts[i]["prompt"] == p
        assert recorder.prompts[i]["response"] == r


def test_message_log_creation(session):
    """Test that a MessageLog instance can be created and saved to the database."""
    log = MessageLog(
        object_name="test_object",
        timestamp="2023-01-01T00:00:00",
        message_log='{"message": "test"}',
    )
    session.add(log)
    session.commit()

    retrieved_log = session.query(MessageLog).first()
    assert retrieved_log.object_name == "test_object"
    assert retrieved_log.timestamp == "2023-01-01T00:00:00"
    assert retrieved_log.message_log == '{"message": "test"}'


def test_upgrade_database(engine):
    """Test that upgrade_database creates the MessageLog table if it doesn't exist."""
    # Drop the table if it exists
    MessageLog.__table__.drop(engine, checkfirst=True)

    # Run upgrade_database
    upgrade_database(engine)

    # Check if the table now exists
    inspector = inspect(engine)
    assert inspector.has_table("message_log")

    # Check if all columns exist
    columns = inspector.get_columns("message_log")
    column_names = [c["name"] for c in columns]
    expected_columns = ["id", "object_name", "timestamp", "message_log"]
    for col in expected_columns:
        assert col in column_names


def test_add_column(engine):
    """Test that add_column adds a new column to an existing table."""
    from sqlalchemy import Column, String

    # Ensure the table exists
    Base.metadata.create_all(engine)

    # Add a new column
    new_column = Column("new_test_column", String)
    with engine.connect() as connection:
        add_column(connection, MessageLog.__tablename__, new_column)

    # Check if the new column exists
    inspector = inspect(engine)
    columns = inspector.get_columns(MessageLog.__tablename__)
    assert any(col["name"] == "new_test_column" for col in columns)


def test_sqlite_log(engine, monkeypatch, tmp_path):
    """Test that sqlite_log correctly logs messages to the database in a temporary directory."""
    # Create a temporary directory for the test database
    temp_db_path = tmp_path / "test_message_log.db"

    # Mock the get_object_name function to return a predictable name
    def mock_get_object_name(obj):
        return "test_object"

    monkeypatch.setattr("llamabot.recorder.get_object_name", mock_get_object_name)

    # Create a mock test object with model_name and temperature attributes
    test_obj = Mock()
    test_obj.model_name = "test_model"
    test_obj.temperature = 0.7

    test_messages = [
        BaseMessage(role="user", content="Hello"),
        BaseMessage(role="assistant", content="Hi there!"),
    ]

    # Call sqlite_log with the temporary database path
    sqlite_log(test_obj, test_messages, db_path=temp_db_path)

    # Create an engine for the temporary database
    temp_engine = create_engine(f"sqlite:///{temp_db_path}")
    Session = sessionmaker(bind=temp_engine)
    session = Session()

    # Query the database to check if the log was saved
    log_entry = session.query(MessageLog).first()

    assert log_entry is not None
    assert log_entry.object_name == "test_object"
    assert "Hello" in log_entry.message_log
    assert "Hi there!" in log_entry.message_log
    assert log_entry.model_name == "test_model"
    assert log_entry.temperature == 0.7

    # Close the session and dispose of the engine
    session.close()
    temp_engine.dispose()

    # Verify that the database file was created
    assert temp_db_path.exists()

    # The temporary directory and its contents will be automatically cleaned up after the test



"""Utilities for llamabot."""

import ast
import fnmatch
import inspect
import sys
from pathlib import Path
from typing import Any, Optional, Union


from pyprojroot import here


def replace_object_in_file(
    source_file_path: str, object_name: str, new_object_definition: str
) -> None:
    """
    Replace an object (function or class)
    in a Python source file with a new object definition.

    .. code-block:: python

        source_file_path = "path/to/your/source_file.py"
        object_name = "function_or_class_name"
        new_object_definition = "def new_function():\\n    pass"
        replace_object_in_file(source_file_path, object_name, new_object_definition)

    :param source_file_path: The path to the source file
        containing the object to be replaced.
    :param object_name: The name of the object (function or class) to be replaced.
    :param new_object_definition: The new object definition as a string.
    :raises SyntaxError: If the source file has invalid Python syntax.
    :raises ValueError: If the object with the specified name
        does not exist in the source file.
    """
    try:
        import astor
    except ImportError:
        raise ImportError(
            "astor is not installed. Please install it with `pip install llamabot[cli]`."
        )

    with open(source_file_path, "r") as source_file:
        source_code = source_file.read()

    try:
        tree = ast.parse(source_code)
    except SyntaxError as e:
        raise SyntaxError(
            "Please ensure the source file has valid Python syntax."
        ) from e

    class ObjectReplacer(ast.NodeTransformer):
        """Internal class used to replace an object in a Python source file."""

        def visit_FunctionDef(
            self, node: ast.FunctionDef
        ) -> Union[ast.FunctionDef, Any]:
            """Replace the object with the specified name.

            :param node: The node to visit.
            :return: The new node.
            """
            if node.name == object_name:
                new_node = ast.parse(new_object_definition).body[0]
                return new_node
            return node

        def visit_ClassDef(self, node: ast.ClassDef) -> Union[ast.ClassDef, Any]:
            """Replace the object with the specified name.

            :param node: The node to visit.
            :return: The new node.
            """
            if node.name == object_name:
                new_node = ast.parse(new_object_definition).body[0]
                return new_node
            return node

    replacer = ObjectReplacer()
    new_tree = replacer.visit(tree)

    if not any(
        isinstance(node, (ast.FunctionDef, ast.ClassDef)) and node.name == object_name
        for node in new_tree.body
    ):
        raise ValueError(
            f"Please ensure the object '{object_name}' exists in the source file."
        )

    new_source_code = astor.to_source(new_tree)

    with open(source_file_path, "w") as source_file:
        source_file.write(new_source_code)


def insert_docstring(source_file_path: str, object_name: str, new_docstring: str):
    """
    Insert a new docstring into a Python source file for a specified object.

    Usage example:

    ```python
    insert_docstring(
        "path/to/source_file.py",
        "function_name",
        "This is the new docstring."
    )
    ```

    :param source_file_path: The path to the source file containing the object.
    :param object_name: The name of the object (function or class)
        to insert the docstring for.
    :param new_docstring: The new docstring to insert.
    :raises SyntaxError: If the source code is not valid Python code.
    :raises ValueError: If the specified object does not exist in the source code.
    """
    try:
        import astor
    except ImportError:
        raise ImportError(
            "astor is not installed. Please install it with `pip install llamabot[cli]`."
        )
    with open(source_file_path, "r") as source_file:
        source_code = source_file.read()

    try:
        tree = ast.parse(source_code)
    except SyntaxError as e:
        raise SyntaxError("Please ensure the source code is valid Python code.") from e

    class DocstringInserter(ast.NodeTransformer):
        """Insert a docstring into a Python source file."""

        def visit_FunctionDef(
            self, node: ast.FunctionDef
        ) -> Union[ast.FunctionDef, None]:
            """Visit a function definition node and insert a docstring
            if the function name matches the specified name.

            :param node: The function definition node.
            :return: The modified function definition node.
            """
            if node.name == object_name:
                docstring = ast.Expr(value=ast.Str(s=new_docstring))
                node.body.insert(0, docstring)
                return node
            return self.generic_visit(node)

        def visit_ClassDef(self, node: ast.ClassDef) -> Union[ast.ClassDef, None]:
            """Visit a class definition node and insert a docstring
            if the class name matches the specified name.

            :param node: The class definition node.
            :return: The modified class definition node.
            """
            if node.name == object_name:
                docstring = ast.Expr(value=ast.Str(s=new_docstring))
                node.body.insert(0, docstring)
                return node
            return self.generic_visit(node)

    inserter = DocstringInserter()
    new_tree = inserter.visit(tree)

    if new_tree is None:
        raise ValueError(
            f"Please ensure the object '{object_name}' exists in the source code."
        )

    new_source_code = astor.to_source(new_tree)

    with open(source_file_path, "w") as source_file:
        source_file.write(new_source_code)


def get_object_source_code(source_file: str, object_name: str) -> Union[str, None]:
    """
    Get the source code of a specific object (function, async function, or class)
    from a source file.

    .. code-block:: python

        source_code = get_object_source_code("example.py", "my_function")

    :param source_file: The path to the source file containing the object.
    :param object_name: The name of the object to get the source code for.
    :return: The source code of the specified object as a string,
        or None if the object is not found.
    :raises SyntaxError: If there is a syntax error in the source file.
    :raises NameError: If the specified object is not found in the source file.
    """
    with open(source_file, "r+") as file:
        source_code = file.read()

    try:
        import astor
    except ImportError:
        raise ImportError(
            "astor is not installed. Please install it with `pip install llamabot[cli]`."
        )

    try:
        parsed_ast = ast.parse(source_code)
    except SyntaxError as e:
        raise SyntaxError(f"Please fix the syntax error in the source file: {e}")

    for node in parsed_ast.body:
        if (
            isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef))
            and node.name == object_name
        ):
            return astor.to_source(node)

    raise NameError(
        f"Please ensure the object '{object_name}' "
        f"exists in the source file '{source_file}'."
    )


def get_function_source(file_path: Union[str, Path], function_name: str) -> str:
    """
    Get the source code of a function from a specified Python file.

    .. code-block:: python

        source_code = get_function_source("path/to/your/file.py", "function_name")

    :param file_path: The path to the Python file containing the function.
    :param function_name: The name of the function to get the source code from.
    :raises FileNotFoundError: If the provided file path is not found.
    :raises ValueError: If the provided file is not a .py file.
    :raises AttributeError: If the specified function is not found in the file.
    :raises TypeError: If the specified name is not a function.
    :return: The source code of the specified function as a string.
    """
    file_path = Path(file_path)
    if not file_path.exists():
        raise FileNotFoundError(
            f"File not found. Please provide a valid file path: {file_path}"
        )

    if not file_path.suffix == ".py":
        raise ValueError(f"Invalid file type. Please provide a .py file: {file_path}")

    sys.path.insert(0, str(file_path.parent))
    module_name = file_path.stem
    module = __import__(module_name)

    function = getattr(module, function_name, None)
    if function is None:
        raise AttributeError(
            f"Function '{function_name}' not found in {file_path}. "
            "Please provide a valid function name."
        )

    if not inspect.isfunction(function):
        raise TypeError(
            f"'{function_name}' is not a function. "
            "Please provide a valid function name."
        )

    return inspect.getsource(function)


def should_ignore_file(file_path: Path, gitignore_patterns: list[str]) -> bool:
    """Check if a file should be ignored.

    :param file_path: The path to the file to check.
    :param gitignore_patterns: A list of patterns to ignore.
    :return: True if the file should be ignored, False otherwise.
    """
    for pattern in gitignore_patterns:
        if fnmatch.fnmatch(file_path.name, pattern):
            return True
    return False


def show_directory_tree(
    directory: str,
    depth: int = float("inf"),
    indent: int = 0,
    ignore_gitignore: bool = True,
    ignore_dirs: list = None,
) -> str:
    """Show directory tree from a directory.

    Example:

    ```python
    from pathlib import Path
    show_directory_tree(
        Path.home(),
        depth=1,
        ignore_gitignore=True,
        ignore_dirs=['.git', '__pycache__']
    )
    ```

    This function prints a directory tree of the specified directory.

    :param directory: The directory to display the tree for.
    :param depth: The maximum depth to display the tree, default is infinity.
    :param indent: The indentation level for the tree, default is 0.
    :param ignore_gitignore: If True, ignores files and folders specified in .gitignore,
        default is True.
    :param ignore_dirs: A list of directory names to ignore, default is None.
    :raises NotADirectoryError: If the specified path is not a directory.
    :returns: A string representation of the directory tree.
    """
    directory_path = Path(directory)
    if not directory_path.is_dir():
        raise NotADirectoryError(
            f"The specified path '{directory}' is not a directory."
        )

    if depth < 0:
        return ""

    files = []
    folders = []

    # Collect files and folders
    for entry in directory_path.iterdir():
        if entry.is_file():
            files.append(entry)
        elif entry.is_dir():
            folders.append(entry)

    # Sort files and folders alphabetically
    files.sort()
    folders.sort()

    # Read gitignore patterns
    gitignore_patterns = []
    gitignore_path = directory_path / ".gitignore"
    if ignore_gitignore and gitignore_path.is_file():
        with gitignore_path.open("r") as gitignore_file:
            gitignore_patterns = gitignore_file.read().splitlines()

    # Prepare the printed text
    printed_text = ""

    # Print files
    for file in files:
        if ignore_gitignore and should_ignore_file(file, gitignore_patterns):
            continue
        printed_text += "|" + "    " * indent + "|-- " + file.name + "\n"

    # Recursively print subdirectories
    for folder in folders:
        if ignore_dirs is not None and any(
            fnmatch.fnmatch(folder.name, pattern) for pattern in ignore_dirs
        ):
            continue
        printed_text += "|" + "    " * indent + "|-- " + folder.name + "/" + "\n"
        if depth > 0:
            printed_text += show_directory_tree(
                folder, depth - 1, indent + 1, ignore_gitignore, ignore_dirs
            )

    return printed_text


def get_git_diff(repo_path: Optional[Union[str, Path]] = None) -> str:
    """Get the git diff of a repository.

    :param repo_path: The path to the git repository.
    :raises ValueError: If the provided path is not a git repository.
    :raises ValueError: If the provided git repository
        has no staged or unstaged changes.
    :return: The git diff as a string, or None if there are no staged changes.
    """
    try:
        from git import GitCommandError, Repo
    except ImportError:
        raise ImportError(
            "git is not installed. Please install it with `pip install llamabot[cli]`."
        )

    if repo_path is None:
        repo_path = here()
    try:
        repo = Repo(repo_path)
    except GitCommandError as e:
        raise ValueError("Please provide a valid path to a git repository.") from e

    if repo.is_dirty():
        if repo.index.diff("HEAD"):
            try:
                diff = repo.git.diff("--cached")
            except GitCommandError as e:
                raise ValueError(
                    "Please ensure that the git repository has staged changes."
                ) from e
        else:
            return ""
    else:
        try:
            diff = repo.git.diff()
        except GitCommandError as e:
            raise ValueError(
                "Please ensure that the git repository has unstaged changes."
            ) from e

    return diff


def get_dependencies(source_file: str, object_name: str) -> list:
    """
    Get a list of paths to other source files
    within the library that the specified object depends on.

    :param source_file: The path to the source file containing the object.
    :param object_name: The name of the object to get the dependencies for.
    :return: A list of paths to other source files.
    """
    dependencies = set()

    # Get the source code of the specified object
    object_source_code = get_object_source_code(source_file, object_name)

    # Collect imports from the object's source code
    parsed_ast = ast.parse(object_source_code)
    for node in ast.walk(parsed_ast):
        if isinstance(node, ast.Import):
            for alias in node.names:
                dependencies.add(alias.name)
        elif isinstance(node, ast.ImportFrom):
            if node.module:
                dependencies.add(node.module)
            for alias in node.names:
                if alias.name != "*":
                    dependencies.add(alias.name)

    # Search for other source files in the library that contain the dependencies
    library_path = Path(source_file).parent
    for path in library_path.glob("**/*.py"):
        if path == source_file:
            continue
        with open(path, "r") as file:
            source_code = file.read()
        parsed_ast = ast.parse(source_code)
        for node in ast.walk(parsed_ast):
            if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef)):
                for decorator in node.decorator_list:
                    if isinstance(decorator, ast.Name) and decorator.id in dependencies:
                        dependencies.remove(decorator.id)
                if node.name in dependencies:
                    dependencies.remove(node.name)
                for child in ast.walk(node):
                    if isinstance(child, ast.Name) and child.id in dependencies:
                        dependencies.remove(child.id)
        if not dependencies:
            break

    return list(map(str, library_path.glob("**/*.py")))



"""Launch the web app to visualize and compare prompts and messages."""

from typing import Optional
import typer
import uvicorn
from pathlib import Path
from llamabot.prompt_manager import find_or_set_db_path
from llamabot.web.app import create_app

app = typer.Typer()


@app.command()
def launch(
    db_path: Optional[Path] = None,
    host: str = typer.Option("0.0.0.0", help="Host to bind the server to"),
    port: int = typer.Option(8000, help="Port to run the server on"),
):
    """Launch the web app to visualize and compare prompts and messages.

    :param db_path: The path to the database to use.
    :param host: The host to bind the server to.
    :param port: The port to run the server on.
    """
    db_path = find_or_set_db_path(db_path)

    fastapi_app = create_app(db_path)
    uvicorn.run(fastapi_app, host=host, port=port)


if __name__ == "__main__":
    app()



"""Generic document store for LlamaBot.

The interface that we need is implemented here:

- append,
- extend,
- retrieve

ChromaDB is a great default choice because of its simplicity and FOSS nature.
Hence we use it by default.
"""

from hashlib import sha256
from pathlib import Path
from typing import Callable, Optional

from tqdm.auto import tqdm

from llamabot.doc_processor import magic_load_doc, split_document


class AbstractDocumentStore:
    """Abstract document store for LlamaBot."""

    def __init__(self):
        raise NotImplementedError()

    def append(self, document: str):
        """Append a document to the store.

        :param document: The document to append.
        :raises NotImplementedError: If the document store is not implemented.
        """
        raise NotImplementedError()

    def extend(self, documents: list[str]):
        """Extend a list of documents to the store.

        :param documents: The documents to append.
        :raises NotImplementedError: If the document store is not implemented.
        """
        raise NotImplementedError()

    def retrieve(self, query: str, n_results: int = 10) -> list[str]:
        """Retrieve a list of documents from the store.

        :param query: The query to make of the documents.
        :param n_results: The number of results to retrieve.
        :raises NotImplementedError: If the document store is not implemented.
        """
        raise NotImplementedError()

    def reset(self):
        """Reset the document store.

        :raises NotImplementedError: If the document store is not implemented.
        """
        raise NotImplementedError()

    def add_documents(
        self,
        document_paths: Path | list[Path],
        chunk_size: int = 2_000,
        chunk_overlap: int = 500,
    ):
        """Add documents to the QueryBot DocumentStore.

        :param document_paths: The document paths to add to the store.
        :param chunk_size: The size of each chunk.
        :param chunk_overlap: The amount of overlap between chunks.
        """
        if isinstance(document_paths, Path):
            document_paths = [document_paths]

        for document_path in tqdm(document_paths):
            document = magic_load_doc(document_path)
            splitted_document = split_document(
                document, chunk_size=chunk_size, chunk_overlap=chunk_overlap
            )
            self.extend(splitted_document)
        self.__post_add_documents__()

    def __post_add_documents__(self):
        """Execute code after adding documents to the store."""
        raise NotImplementedError()


class ChromaDBDocStore(AbstractDocumentStore):
    """A document store for LlamaBot that wraps around ChromaDB."""

    def __init__(
        self,
        collection_name: str,
        storage_path: Path = Path.home() / ".llamabot" / "chroma.db",
    ):
        try:
            import chromadb
        except ImportError:
            raise ImportError(
                "ChromaDB is required for ChromaDBDocStore. "
                "Please `pip install llamabot[rag]` to use the ChromaDB document store."
            )

        client = chromadb.PersistentClient(path=str(storage_path))
        collection = client.create_collection(collection_name, get_or_create=True)

        self.storage_path = storage_path
        self.client = client
        self.collection = collection
        self.collection_name = collection_name
        self.existing_records = collection.get()

    def __post_add_documents__(self):
        """Execute code after adding documents to the store."""
        pass

    def append(
        self,
        document: str,
        metadata: Optional[dict] = None,
        embedding: Optional[list[float]] = None,
    ):
        """Append a document to the store.

        By default, the documents will be automatically embedded
        using the default embedder that is set in the ChromaDB client.
        However, there may be cases where we want to pre-compute the embeddings
        and pass them into the store, e.g. when adding documents in parallel.
        In this case, we can pass the pre-computed embeddings into the store.
        Use `append` to do single adds, and `extend` to do bulk adds.

        :param document: The document to append.
        :param metadata: The metadata to append.
        :param embedding: The embedding to append, optional
        """
        doc_id = sha256(document.encode()).hexdigest()
        self.collection.add(
            documents=document,
            ids=doc_id,
            metadatas=metadata,
            embeddings=embedding,
        )

    def extend(
        self,
        documents: list[str],
        metadatas: Optional[list[dict]] = None,
        embeddings: Optional[list[list[float]]] = None,
    ):
        """Extend the document store.

        This is effectively a bulk add of documents.
        See the docstring for `append` for more details.

        :param documents: Iterable of documents.
        :param metadatas: Iterable of metadatas.
        :param embeddings: Iterable of pre-computed embeddings.
        """
        # Compute doc_id
        ids = [sha256(doc.encode()).hexdigest() for doc in documents]

        # Check that the lengths of the lists are the same
        if metadatas is not None and len(documents) != len(metadatas):
            raise ValueError(
                "The lengths of the documents and metadatas must be the same."
            )
        if embeddings is not None and len(documents) != len(embeddings):
            raise ValueError(
                "The lengths of the documents and embeddings must be the same."
            )

        self.collection.add(
            ids=ids,
            documents=documents,
            metadatas=metadatas,
            embeddings=embeddings,
        )

    def retrieve(self, query: str, n_results: int = 10) -> list[str]:
        """Retrieve documents from the store.

        :param query: The query to use to retrieve documents.
        """
        try:
            import chromadb
        except ImportError:
            raise ImportError(
                "ChromaDB is required for ChromaDBDocStore. "
                "Please `pip install llamabot[rag]` to use the ChromaDB document store."
            )

        # Use Vectordb to get documents.
        results: chromadb.QueryResult = self.collection.query(
            query_texts=query, n_results=n_results
        )
        vectordb_documents: list[str] = results["documents"][0]
        return vectordb_documents

        # Return the union of the retrieved documents
        # union = set(vectordb_documents).union(bm25_documents)
        # return list(union)

    def reset(self):
        """Reset the document store."""
        self.client.delete_collection(self.collection_name)
        self.collection = self.client.create_collection(
            self.collection_name, get_or_create=True
        )


class LanceDBDocStore(AbstractDocumentStore):
    """A document store for LlamaBot that wraps around LanceDB."""

    def __init__(
        self,
        table_name: str,
        storage_path: Path = Path.home() / ".llamabot" / "lancedb",
        auto_create_fts_index: bool = True,
    ):
        try:
            import lancedb
            from lancedb.embeddings import EmbeddingFunctionRegistry, get_registry
            from lancedb.pydantic import LanceModel, Vector
        except ImportError:
            raise ImportError(
                "LanceDB is required for LanceDBDocStore. "
                "Please `pip install llamabot[rag]` to use the LanceDB document store."
            )

        registry: EmbeddingFunctionRegistry = get_registry()
        func: Callable = registry.get(name="sentence-transformers").create()

        class DocstoreEntry(LanceModel):
            """LanceDB DocumentStore Entry."""

            document: str = func.SourceField()
            vector: Vector(func.ndims()) = func.VectorField()

        self.schema = DocstoreEntry
        self.table_name = table_name
        storage_path.mkdir(parents=True, exist_ok=True)  # Ensure storage path exists
        self.db = lancedb.connect(storage_path)

        try:
            self.table = self.db.open_table(table_name)
        except Exception:
            self.table = self.db.create_table(table_name, schema=self.schema)

        try:
            self.existing_records = [
                item.document
                for item in self.table.search().limit(None).to_pydantic(self.schema)
            ]
        except ValueError:
            self.existing_records = []

        if auto_create_fts_index:
            self.table.create_fts_index(field_names=["document"], replace=True)

    def __contains__(self, other: str) -> bool:
        """Returns boolean whether the 'other' document is in the store.

        :param other: The document to search for.
        :return: True if the document is in the store, False otherwise.
        """
        all_items = self.table.search().limit(None).to_pydantic(self.schema)
        texts = set([item.document for item in all_items])
        return other in texts

    def append(self, document: str):
        """Append a document to the store.

        :param document: The document to append.
        """
        # Avoid duplication of documents in LanceDB.
        if document not in self.existing_records:
            self.table.add([{"document": document}])
            self.existing_records.append(document)

    def extend(self, documents: list[str]):
        """Extend a list of documents to the store.

        :param documents: The documents to append.
        """
        # self.table.add(documents)
        for doc in tqdm(documents):
            self.append(doc)

    def retrieve(self, query: str, n_results: int = 10) -> list[str]:
        """Retrieve a list of documents from the store.

        :param query: The query to use to retrieve documents.
        :param n_results: The number of results to retrieve.
        :return: A list of documents.
        """
        results: list = (
            self.table.search(query, query_type="hybrid")
            .limit(n_results)
            .to_pydantic(self.schema)
        )
        return [r.document for r in results]

    def reset(self):
        """Reset the document store."""
        self.db.drop_table(self.table_name)
        self.table = self.db.create_table(self.table_name, schema=self.schema)

    def __post_add_documents__(self):
        """Execute code after adding documents to the store."""
        self.table.create_fts_index("document", replace=True)


class BM25DocStore(AbstractDocumentStore):
    """In-memory BM25-based retrieval document store."""

    def __init__(self):
        self.documents: list = []

    def __post_add_documents__(self):
        """Execute code after adding documents to the store."""
        pass

    def append(self, document: str):
        """Append a document to the store.

        :param document: The document to append.
        """
        self.documents.append(document)

    def extend(self, documents: list[str]):
        """Extend a list of documents to the store.

        :param documents: The documents to append.
        """
        self.documents.extend(documents)

    def retrieve(self, query: str, n_results: int = 10) -> list[str]:
        """Retrieve documents from the store.

        :param query: The query to use to retrieve documents.
        :param n_results: The number of results to retrieve.
        :return: A list of documents.
        """

        try:
            from rank_bm25 import BM25Okapi
        except ImportError:
            raise ImportError(
                "rank_bm25 is required for BM25DocStore. "
                "Please install it with `pip install llamabot[rag]`"
            )

        # Use BM25 to get documents.
        tokenized_docs = [doc.split() for doc in self.documents]
        search_engine = BM25Okapi(tokenized_docs)
        docs: list[str] = search_engine.get_top_n(
            query.split(), self.documents, n=n_results
        )
        return docs

    def reset(self):
        """Reset the document store."""
        self.documents = []



"""Tests for the experiments module.

This module contains tests for the experiment tracking functionality,
including metric recording, message logging, and prompt tracking.
"""

from pathlib import Path

import pytest
from sqlalchemy import create_engine, select
from sqlalchemy.orm import sessionmaker

from llamabot import prompt, Experiment, metric
from llamabot.experiments import ExperimentRun, Base
from llamabot.recorder import sqlite_log
from llamabot.components.messages import BaseMessage
from llamabot.bot.simplebot import SimpleBot


@pytest.fixture
def db_path(tmp_path) -> Path:
    """Create a temporary database path.

    :param tmp_path: pytest fixture for temporary directory
    :return: Path to test database
    """
    return tmp_path / "test.db"


@pytest.fixture
def engine(db_path):
    """Create a test database engine.

    :param db_path: Path to test database
    :return: SQLAlchemy engine
    """
    engine = create_engine(f"sqlite:///{db_path}")
    Base.metadata.create_all(engine)
    return engine


@pytest.fixture
def session(engine):
    """Create a test database session.

    :param engine: SQLAlchemy engine
    :return: SQLAlchemy session
    """
    Session = sessionmaker(bind=engine)
    return Session()


def test_metric_recording(db_path):
    """Test that metrics are properly recorded in experiment runs."""

    @metric
    def test_metric(x: int) -> float:
        """Test metric that returns a constant.

        :param x: Input value (unused)
        :return: Constant value of 42.0
        """
        return 42.0

    with Experiment("test_metrics", db_path=db_path) as exp:
        value = test_metric(0)
        assert value == 42.0

        # Check that the metric was recorded
        assert exp.run is not None
        assert "test_metric" in exp.run.run_data["metrics"]
        assert exp.run.run_data["metrics"]["test_metric"]["value"] == 42.0


def test_message_logging(db_path):
    """Test that message logs are properly linked to experiment runs."""

    # Create a SimpleBot instance for testing
    test_bot = SimpleBot(system_prompt="Test system prompt")

    with Experiment("test_messages", db_path=db_path) as exp:
        messages = [
            BaseMessage(role="user", content="Hello"),
            BaseMessage(role="assistant", content="Hi there!"),
        ]

        # Log messages using the test_bot instead of object()
        log_id = sqlite_log(test_bot, messages, db_path)

        # Check that the message log was linked
        assert exp.run is not None
        assert log_id in exp.run.run_data["message_log_ids"]


def test_prompt_tracking(db_path):
    """Test that prompts are properly tracked in experiment runs."""

    @prompt("system")
    def test_prompt(x: str) -> str:
        """Test prompt with parameter {{ x }}."""
        return f"Test prompt with {x}"

    with Experiment("test_prompts", db_path=db_path) as exp:
        _ = test_prompt("hello")

        # Check that the prompt was tracked
        assert exp.run is not None
        assert len(exp.run.run_data["prompts"]) == 1
        prompt_info = exp.run.run_data["prompts"][0]
        assert "hash" in prompt_info
        assert isinstance(prompt_info["hash"], str)


def test_experiment_metadata(db_path):
    """Test that experiment metadata is properly stored."""
    metadata = {"test_key": "test_value"}

    with Experiment("test_metadata", metadata=metadata, db_path=db_path) as exp:
        assert exp.run is not None
        assert exp.run.run_data["metadata"] == metadata


def test_multiple_runs(db_path):
    """Test that multiple runs are properly tracked."""
    engine = create_engine(f"sqlite:///{db_path}")
    Session = sessionmaker(bind=engine)

    # First run
    with Experiment("multiple_runs", db_path=db_path):
        pass

    # Second run
    with Experiment("multiple_runs", db_path=db_path):
        pass

    # Check that both runs were recorded
    session = Session()
    runs = (
        session.execute(
            select(ExperimentRun).where(
                ExperimentRun.experiment_name == "multiple_runs"
            )
        )
        .scalars()
        .all()
    )

    assert len(runs) == 2
    session.close()


def test_experiment_cleanup(db_path):
    """Test that experiment resources are properly cleaned up."""
    exp = Experiment("cleanup_test", db_path=db_path)

    with exp:
        assert exp.run is not None
        assert exp.session is not None

    # After context exit
    assert exp.run is None

    # Try to use the session (should raise an error)
    with pytest.raises(Exception):
        exp.session.query(ExperimentRun).all()


def test_experiment_exception_handling(db_path):
    """Test that experiments handle exceptions properly."""
    exp = Experiment("exception_test", db_path=db_path)

    with pytest.raises(ValueError):
        with exp:
            raise ValueError("Test exception")

    # Session should be closed
    with pytest.raises(Exception):
        exp.session.query(ExperimentRun).all()


def test_run_data_structure(db_path):
    """Test that the run_data JSON structure is correct."""
    with Experiment("structure_test", db_path=db_path) as exp:
        assert exp.run is not None
        run_data = exp.run.run_data

        # Check structure
        assert "metadata" in run_data
        assert "message_log_ids" in run_data
        assert "prompts" in run_data
        assert "metrics" in run_data

        # Check types
        assert isinstance(run_data["metadata"], dict)
        assert isinstance(run_data["message_log_ids"], list)
        assert isinstance(run_data["prompts"], list)
        assert isinstance(run_data["metrics"], dict)


def test_concurrent_experiments(db_path):
    """Test that nested experiment contexts work properly."""
    # Create a SimpleBot instance for testing
    test_bot = SimpleBot(system_prompt="Test system prompt")

    with Experiment("outer", db_path=db_path) as outer_exp:
        with Experiment("inner", db_path=db_path) as inner_exp:
            assert outer_exp.run is not None
            assert inner_exp.run is not None

            # Log something in inner experiment
            messages = [
                BaseMessage(role="user", content="Hello"),
                BaseMessage(role="assistant", content="Hi!"),
            ]
            log_id = sqlite_log(test_bot, messages, db_path)

            # Check that the log is associated with the inner experiment
            assert log_id in inner_exp.run.run_data["message_log_ids"]
            assert log_id not in outer_exp.run.run_data["message_log_ids"]



"""Code for document preprocessing."""

from pathlib import Path
from typing import List

from pdfminer.high_level import extract_text


def pdf_loader(path: Path) -> str:
    """Load a PDF file from a file path.

    :param path: Path to the PDF file to be loaded.
    :return: Text extracted from the PDF file.
    """
    return extract_text(str(path))


def plaintext_loader(path: Path) -> str:
    """Load a plain text file from a file path.

    :param path: Path to the plain text file to be loaded.
    :return: Text from the plain text file.
    """
    with open(path, "r+") as f:
        return f.read()


EXTENSION_LOADER_MAPPING = {
    ".pdf": pdf_loader,
    # ".docx": "DocxReader",
    # ".pptx": "PptxReader",
    # ".xlsx": "PandasExcelReader",
}


def magic_load_doc(file_path: Path) -> str:
    """Load a document from a file.

    This function is used to magically load a document from a file.
    The file extension is inferred from the file path.
    If the file extension is not recognized, it is assumed to be a plain text file.

    :param file_path: Path to the file to be loaded.
    :return: A list of documents.
    """
    loader = EXTENSION_LOADER_MAPPING.get(Path(file_path).suffix, plaintext_loader)
    return loader(file_path)


def split_document(
    doc: str,
    chunk_size: int = 2000,
    chunk_overlap: int = 0,
) -> List[str]:
    """Split a document into sub-documents using token text splitter.

    This function is used to split a document
    into sub-documents using token text splitter,
    such that each sub-document has a maximum length of `chunk_size` tokens
    with `chunk_overlap` tokens overlap.

    :param doc: Document to be split.
    :param chunk_size: Maximum length of each sub-document.
    :param chunk_overlap: Number of tokens to overlap between each sub-document.
    :raises ValueError: If `chunk_overlap` is negative.
    :return: A list of sub-documents.
    """
    if chunk_overlap < 0:
        raise ValueError("chunk_overlap must be non-negative.")
    if chunk_size < 0:
        raise ValueError("chunk_size must be non-negative.")

    # Split the document into sub-documents with chunk size of `chunk_size` characters
    # and overlap of `chunk_overlap` characters.
    sub_texts = []
    for i in range(0, len(doc), chunk_size - chunk_overlap):
        sub_texts.append(doc[i : i + chunk_size])
    return sub_texts



"""Completer class for paper titles."""

from prompt_toolkit.completion import Completer, Completion


class PaperTitleCompleter(Completer):
    """Completer class for paper titles.

    :param paper_titles: A list of paper titles to choose from."""

    def __init__(self, paper_titles):
        self.paper_titles = paper_titles

    def get_completions(self, document, complete_event) -> list:
        """Get completions for the given document.

        :param document: The document to get completions for.
        :param complete_event: The event that triggered the completion.
        :return: A list of completions. Naomi mama
        """
        text = document.text_before_cursor
        completions = []

        if text:
            for title in self.paper_titles:
                if all(word.lower() in title.lower() for word in text.split()):
                    completions.append(Completion(title, start_position=-len(text)))

        return completions



"""Prompts and bots for validating outputs."""

import json

from llamabot import SimpleBot
from llamabot.prompt_manager import prompt


@prompt(role="system")
def output_formatter_sysprompt():
    """You are an output formatter.

    You will be presented with a text that should be turned into a JSON.
    Please convert it to a JSON.
    Ensure that the keys are always in snake case.
    Use as few words as possible for the keys.
    """


def output_formatter() -> SimpleBot:
    """Return the output formatter bot.

    :return: The output formatter bot.
    """
    return SimpleBot(
        output_formatter_sysprompt(),
        temperature=0.2,
    )


def coerce_dict(bot_response: str) -> dict:
    """Coerce the bot_response to be a dictionary.

    :param bot_response: The bot response to be coerced.
    :return: The bot response as a dictionary.
    :raises ValueError: If the bot response cannot be coerced to a dictionary.
    """
    formatter: SimpleBot = output_formatter()

    max_retries = 5
    num_retries = 0
    is_dict = False
    while not is_dict and num_retries < max_retries:
        try:
            parsed_dictionary = json.loads(bot_response)
            is_dict = True
            return parsed_dictionary
        except Exception:
            pass
        bot_response = formatter(bot_response).content
        num_retries += 1
    raise ValueError(f"Could not coerce to dictionary after {max_retries} tries.")



"""Utility functions."""

from pathlib import Path
from typing import Optional

from pyprojroot import here


def get_object_name(obj):
    """
    Get the name of the object as it's defined in the current namespace.

    :param obj: The object whose name we want to find.
    :return: The name of the object as a string, or None if not found.
    """
    for name, value in globals().items():
        if value is obj:
            return name
    return None


def find_or_set_db_path(db_path: Optional[Path] = None) -> Path:
    """Find or set the database path for message logging.

    If no path is provided, attempts to create the database in the current project root.
    Falls back to user's home directory if project root cannot be determined.

    :param db_path: Optional path to the database file. If None, uses default locations.
    :return: Path to the database file.
    """
    if db_path is None:
        try:
            db_path = here() / "message_log.db"
        except Exception:
            db_path = Path.home() / ".llamabot" / "message_log.db"
            db_path.parent.mkdir(parents=True, exist_ok=True)
    return db_path



"""A module implementing an agent-based bot that can execute a sequence of tools.

This module provides the AgentBot class, which combines language model capabilities
with the ability to execute a sequence of tools based on user input.
The bot uses a decision-making system to determine which tools to call
and in what order, making it suitable for complex, multi-step tasks.
"""

import hashlib
import json
from datetime import datetime
from typing import Any, Callable, Dict, List, Optional, Union
from uuid import uuid4

from pydantic import BaseModel, Field

from llamabot.bot.structuredbot import StructuredBot
from llamabot.components.messages import AIMessage, BaseMessage, system, user
from llamabot.components.sandbox import ScriptExecutor, ScriptMetadata
from llamabot.components.tools import tool
from llamabot.config import default_language_model
import requests
from bs4 import BeautifulSoup


@tool
def search_internet(
    search_term: str, max_results: int = 10, backend: str = "lite"
) -> Dict[str, str]:
    """Search the internet for a given term and get webpage contents.

    :param search_term: The search term to look up
    :param max_results: Maximum number of search results to return
    :param backend: The DuckDuckGo backend to use
    :return: Dictionary mapping URLs to markdown-formatted webpage contents
    """
    try:
        from duckduckgo_search import DDGS
    except ImportError:
        raise ImportError(
            "The Python package `duckduckgo_search` cannot be found. Please install it using `pip install llamabot[agent]`."
        )

    try:
        from markdownify import markdownify as md
    except ImportError:
        raise ImportError(
            "The Python package `markdownify` cannot be found. Please install it using `pip install llamabot[agent]`."
        )

    ddgs = DDGS()
    results = ddgs.text(search_term, max_results=max_results, backend=backend)
    webpage_contents = {}
    headers = {
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36",
        "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8",
        "Accept-Language": "en-US,en;q=0.5",
        "Accept-Encoding": "gzip, deflate, br",
        "DNT": "1",
        "Connection": "keep-alive",
    }
    for result in results:
        response = requests.get(result["href"], headers=headers, timeout=10)
        soup = BeautifulSoup(response.text, "html.parser")
        webpage_contents[result["href"]] = md(soup.get_text())

    return webpage_contents


@tool
def agent_finish(message: Any) -> str:
    """Tool to indicate that the agent has finished processing the user's message."""
    try:
        return str(message)
    except Exception:
        return repr(message)


@tool
def return_error(message: Any) -> str:
    """Tool to indicate that the agent has encountered an error.

    :param message: The error message or exception to raise
    :raises Exception: Always raises the provided error message as an exception
    """
    raise Exception(str(message))


@tool
def write_and_execute_script(
    code: str,
    dependencies_str: Optional[str] = None,
    python_version: str = ">=3.11",
    timeout: int = 30,
) -> Dict[str, Any]:
    """Write and execute a Python script in a secure sandbox.

    Dependencies should be specified as a comma-separated string, e.g. "requests,beautifulsoup4".

    Script output will be captured from stdout. Use print() to output results.

    :param code: The Python code to execute
    :param dependencies_str: Comma-separated string of pip dependencies
    :param python_version: Python version requirement
    :param timeout: Execution timeout in seconds
    :return: Dictionary containing script execution results
    """
    # Parse dependencies string into list
    dependencies = list(
        dep.strip()
        for dep in (dependencies_str or "").split(",")
        if dep.strip()  # Filter out empty strings
    )
    # Modify the Python version if it doesn't have a version specifier
    if not any(python_version.startswith(op) for op in (">=", "<=", "==", "~=")):
        python_version = f">={python_version}"

    # Create metadata
    metadata = ScriptMetadata(
        requires_python=python_version,
        dependencies=dependencies,
        auth=str(uuid4()),  # Generate unique ID for this execution
        timestamp=datetime.now(),
    )

    # Initialize executor
    executor = ScriptExecutor()

    # Write and run script
    script_path = executor.write_script(code, metadata)
    result = executor.run_script(script_path, timeout)

    # Return structured output
    return {
        "stdout": result["stdout"].strip(),
        "stderr": result["stderr"].strip(),
        "status": result["status"],
    }


class CachedResult(BaseModel):
    """Model for storing cached results from tool executions.

    :param tool_name: Name of the tool that generated the result
    :param tool_arguments: Dictionary of arguments used to call the tool
    :param result: The actual result value
    :param timestamp: When the result was generated
    :param hash_key: SHA256 hash of the stringified result
    """

    tool_name: str
    tool_arguments: Dict[str, Any]
    result: Any
    timestamp: datetime = Field(default_factory=datetime.now)
    hash_key: str = Field(
        ...,
        description="SHA256 hash of the stringified result. Length 8 characters.",
    )


def hash_result(result: Any) -> str:
    """Generate a SHA256 hash for a result value.

    :param result: Any result value that can be converted to string
    :return: First 8 characters of the SHA256 hash of the stringified result
    """
    # Convert result to a consistent string representation
    result_str = json.dumps(result, sort_keys=True, default=str)
    return hashlib.sha256(result_str.encode()).hexdigest()[:8]


class ToolArgument(BaseModel):
    """Pydantic model for tool arguments.

    :param name: Name of the argument
    :param value: Value of the argument, can be None if coming from cache
    """

    name: str = Field(..., description="Name of the argument.")
    value: Optional[
        Union[str, int, float, bool, dict, List[Union[str, int, float, bool, dict]]]
    ] = Field(
        ...,
        description="Argument value. None if coming from cache.",
    )


class CachedArguments(BaseModel):
    """Pydantic model for cached argument references.

    :param arg_name: Name of the argument to populate from cache
    :param hash_key: Hash key of the cached result to use
    """

    arg_name: str = Field(
        ..., description="Name of the argument to populate from cache."
    )
    hash_key: str = Field(
        ...,
        description="Hash key of the cached result to use. It is a sha256 hash of the result.",
    )


class ToolToCall(BaseModel):
    """Pydantic model representing a single tool to be called by the agent.

    :param tool_name: Name of the tool to call
    :param tool_args: List of tool arguments and their values
    :param use_cached_results: List of argument names and hash keys
        that should be used as input for those arguments
    """

    tool_name: str = Field(..., description="The tool to call.")
    tool_args: list[ToolArgument] = Field(
        ...,
        description="Arguments to pass to the tool. Use None for arguments that should come from cache.",
    )
    use_cached_results: list[CachedArguments] = Field(
        ...,
        description="List of argument names and result hash keys to use as input.",
    )


class AgentBot:
    """A bot that uses an agent to process messages and execute tools.

    Additional attributes:
    :param memory: Dictionary storing cached results from tool executions,
                  indexed by SHA256 hash of the result
    """

    def __init__(
        self,
        system_prompt: str,
        temperature=0.0,
        model_name=default_language_model(),
        stream_target: str = "stdout",
        api_key: Optional[str] = None,
        mock_response: Optional[str] = None,
        functions: Optional[list[Callable]] = None,
        **completion_kwargs,
    ):
        self.memory: Dict[str, CachedResult] = {}

        # Update decision bot prompt to include information about using cached results
        self.decision_bot = StructuredBot(
            pydantic_model=ToolToCall,
            system_prompt=system(
                system_prompt,
                "Given the following message and available tools, "
                "pick the next tool that you need to call "
                "and the arguments that you need for it. "
                "You have access to previously cached results - use them when appropriate "
                "by specifying their hash key in use_cached_results. "
                "Any arguments that you do not know ahead of time, just leave blank. "
                "Only call tools that are relevant to the user's message. "
                "If the task is complete, use the agent_finish tool.",
            ),
            model_name=model_name,
            temperature=temperature,
            stream_target=stream_target,
            api_key=api_key,
            mock_response=mock_response,
            **completion_kwargs,
        )

        functions = [
            agent_finish,
            return_error,
        ] + (functions or [])
        self.functions = functions
        self.tools = {func.__name__: func for func in self.functions}

    def _store_result(
        self, tool_name: str, result: Any, tool_arguments: Dict[str, Any]
    ) -> str:
        """Store a tool execution result in memory.

        :param tool_name: Name of the tool that generated the result
        :param tool_arguments: Dictionary of arguments used to call the tool
        :param result: The result to store
        :return: The hash key under which the result is stored
        """
        result_hash = hash_result(result)

        # If this exact result is already cached, return its hash
        if result_hash in self.memory:
            return result_hash

        # Store new result indexed by its hash
        self.memory[result_hash] = CachedResult(
            tool_name=tool_name,
            tool_arguments=tool_arguments,
            result=result,
            hash_key=result_hash,
        )
        return result_hash

    def __call__(
        self,
        *messages: Union[str, BaseMessage, List[Union[str, BaseMessage]]],
        max_iterations: int = 10,
    ) -> AIMessage:
        """Process messages and execute the appropriate sequence of tools.

        :param messages: One or more messages to process
        :param max_iterations: Maximum number of iterations to run
        :return: The final response as an AIMessage
        :raises RuntimeError: If max iterations is reached
        """
        results = []
        execution_history = []
        iteration = 0

        while iteration < max_iterations:
            iteration += 1
            next_tool = self.decision_bot(
                user(
                    "This is the problem you need to solve: ",
                    *messages,
                    "Here are the available tools: ",
                    *[json.dumps(func.json_schema) for func in self.functions],
                    "You are allowed to write as many different scripts as you need to solve the problem. ",
                    "If you encounter an error, try taking a step back and thinking about the problem ",
                    "before calling the next tool."
                    "Here are previously cached results you can use: ",
                    *[
                        f"{k}: {v.tool_name} result from {v.timestamp}"
                        for k, v in self.memory.items()
                    ],
                    "Here is the execution history for reference: ",
                    *execution_history,
                    "If you have the answer, use the agent_finish tool to finish the task and write an answer to the user.",
                )
            )

            # Prepare arguments from tool_args
            args = {}
            for arg in next_tool.tool_args:
                if arg.value is not None:
                    args[arg.name] = arg.value

            try:
                # Add cached results where specified
                for cached_arg in next_tool.use_cached_results:
                    if cached_arg.hash_key not in self.memory:
                        raise ValueError(
                            f"Cached result {cached_arg.hash_key} not found"
                        )
                    args[cached_arg.arg_name] = self.memory[cached_arg.hash_key].result

                tool_name = next_tool.tool_name
                result = self.tools[tool_name](**args)

                if tool_name == "write_and_execute_script":
                    print("---SCRIPT EXECUTION RESULT---")
                    print(result["stdout"])
                    print("---SCRIPT EXECUTION RESULT---")
                    result = result["stdout"]
                result_id = self._store_result(tool_name, result, args)

                # Log result - only show tool name, args, and result ID for conciseness
                result_str = f"{tool_name}({args}) -> {result_id}, {str(result)[0:100]}"
                results.append(result)
                execution_history.append(result_str)

                # Handle special tool cases
                if tool_name == "agent_finish":
                    return AIMessage(content=str(result))
                if tool_name == "return_error":
                    raise Exception(result)

            except Exception as e:
                error_msg = f"Error calling {next_tool.tool_name}: {str(e)}"
                print()
                print(error_msg)
                execution_history.append(error_msg)
                if next_tool.tool_name == "return_error":
                    raise e
                continue

        raise RuntimeError(f"Agent exceeded maximum iterations ({max_iterations})")



"""Class definition for SimpleBot."""

import contextvars
from typing import Generator, Optional, Union


from llamabot.components.messages import (
    AIMessage,
    HumanMessage,
    SystemMessage,
    BaseMessage,
    process_messages,
)
from llamabot.recorder import autorecord, sqlite_log
from llamabot.config import default_language_model
from pydantic import BaseModel

prompt_recorder_var = contextvars.ContextVar("prompt_recorder")


class SimpleBot:
    """Simple Bot that is primed with a system prompt, accepts a human message,
    and sends back a single response.

    This bot does not retain chat history.

    :param system_prompt: The system prompt to use.
    :param temperature: The model temperature to use.
        See https://platform.openai.com/docs/api-reference/completions/create#completions/create-temperature
        for more information.
    :param model_name: The name of the model to use.
    :param stream_target: The target to stream the response to.
        Should be one of ("stdout", "panel", "api").
    :param json_mode: Whether to print debug messages.
    :param api_key: The OpenAI API key to use.
    :param mock_response: A mock response to use, for testing purposes only.
    :param completion_kwargs: Additional keyword arguments to pass to the completion function.
    """

    def __init__(
        self,
        system_prompt: str,
        temperature=0.0,
        model_name=default_language_model(),
        stream_target: str = "stdout",
        json_mode: bool = False,
        api_key: Optional[str] = None,
        mock_response: Optional[str] = None,
        **completion_kwargs,
    ):
        # Validate `stream_target.
        if stream_target not in ("stdout", "panel", "api", "none"):
            raise ValueError(
                f"stream_target must be one of ('stdout', 'panel', 'api', 'none'), got {stream_target}."
            )

        self.system_prompt = system_prompt
        if isinstance(system_prompt, str):
            self.system_prompt = SystemMessage(content=system_prompt)

        self.model_name = model_name
        self.stream_target = stream_target
        self.temperature = temperature
        self.json_mode = json_mode
        self.api_key = api_key
        self.mock_response = mock_response
        self.completion_kwargs = completion_kwargs

        # Set special cases for for o1 models.
        if model_name in ["o1-preview", "o1-mini"]:
            self.system_prompt = HumanMessage(
                content=system_prompt.content, prompt_hash=system_prompt.prompt_hash
            )
            self.temperature = 1.0
            self.stream_target = "none"

    def __call__(
        self, *human_messages: Union[str, BaseMessage, list[Union[str, BaseMessage]]]
    ) -> Union[AIMessage, Generator]:
        """Call the SimpleBot.

        :param human_messages: One or more human messages to use, or lists of messages.
        :return: The response to the human messages, primed by the system prompt.
        """
        processed_messages = process_messages(human_messages)
        messages = [self.system_prompt] + processed_messages
        match self.stream_target:
            case "stdout":
                return self.stream_stdout(messages)
            case "panel":
                return self.stream_panel(messages)
            case "api":
                return self.stream_api(messages)
            case "none":
                return self.stream_none(messages)
        return AIMessage(content="")

    # @cache.memoize(ignore={0})
    def stream_stdout(self, messages: list[BaseMessage]) -> AIMessage:
        """Stream the response to stdout.

        :param messages: A list of messages.
        """
        response = _make_response(self, messages)
        message = ""
        for chunk in response:
            delta = chunk.choices[0].delta["content"]
            if delta is not None:
                print(delta, end="")
                message += delta
        response_message = AIMessage(content=message)
        autorecord(messages[-1].content, response_message.content)
        sqlite_log(self, messages + [response_message])
        return response_message

    def stream_panel(self, messages: list[BaseMessage]) -> Generator:
        """Stream the response to a Panel app.

        :param messages: A list of messages.
        """
        response = _make_response(self, messages)
        response_message = ""
        for chunk in response:
            delta = chunk.choices[0].delta["content"]
            if delta is not None:
                response_message += delta
                yield response_message
        autorecord(messages[-1].content, response_message)
        sqlite_log(self, messages + [AIMessage(content=response_message)])

    # @cache.memoize(ignore={0})
    def stream_none(self, messages: list[BaseMessage]) -> AIMessage:
        """Stream the response to None.

        :param messages: A list of messages.
        """
        response = _make_response(self, messages, stream=False)
        response_message = AIMessage(
            content=response.choices[0].message.content.strip()
        )
        autorecord(messages[-1].content, response_message.content)
        sqlite_log(self, messages + [response_message])
        return response_message

    def stream_api(self, messages: list[BaseMessage]) -> Generator:
        """Stream the response to an API.

        :param messages: A list of messages.
        """
        response = _make_response(self, messages)
        response_message = ""
        for chunk in response:
            delta = chunk.choices[0].delta["content"]
            if delta is not None:
                response_message += delta
                yield delta
        autorecord(messages[-1].content, response_message)
        sqlite_log(self, messages + [AIMessage(content=response_message)])

    # @cache.memoize(ignore={0})
    def generate_response(self, messages: list[BaseMessage]) -> AIMessage:
        """Generate a response from the given messages.

        :param messages: A list of messages.
        :return: The response to the messages.
        """
        response = _make_response(self, messages)
        response_message = AIMessage(
            content=response.choices[0].message.content.strip()
        )
        autorecord(messages[-1].content, response_message.content)
        sqlite_log(self, messages + [response_message])
        return response_message

    def stream_response(self, messages: list[BaseMessage]):
        """Stream the response from the given messages.

        This is intended to be used with Panel's ChatInterface as part of the callback.

        :param messages: A list of messages.
        :return: A generator that yields the response.
        """
        response = _make_response(self, messages)
        message = ""
        for chunk in response:
            delta = chunk.choices[0].delta.content
            if delta is not None:
                message += delta
                print(delta, end="")
                yield message
        print()


def _make_response(bot: SimpleBot, messages: list[BaseMessage], stream: bool = True):
    """Make a response from the given messages.

    :param bot: A SimpleBot
    :param messages: A list of Messages.
    :return: A response object.
    """
    from litellm import completion

    messages_dumped: list[dict] = [
        {k: v for k, v in m.model_dump().items() if k in ("role", "content")}
        for m in messages
    ]
    completion_kwargs = dict(
        model=bot.model_name,
        messages=messages_dumped,
        temperature=bot.temperature,
        stream=stream,
    )
    completion_kwargs.update(bot.completion_kwargs)

    if bot.mock_response:
        completion_kwargs["mock_response"] = bot.mock_response
    if bot.json_mode:
        # Check if bot has pydantic_model attribute and it's a BaseModel
        if not hasattr(bot, "pydantic_model"):
            raise ValueError(
                "Please set a pydantic_model for this bot to use JSON mode!"
            )
        if not issubclass(getattr(bot, "pydantic_model"), BaseModel):
            raise ValueError("pydantic_model must be a Pydantic BaseModel class")

        completion_kwargs["response_format"] = getattr(bot, "pydantic_model")
    if bot.api_key:
        completion_kwargs["api_key"] = bot.api_key
    return completion(**completion_kwargs)



"""Tools for AgentBots to call on.

The design of a tool is as follows:

1. It is a function that is callable.
2. It is decorated with @tool.
3. Being decorated by @tool,
   it will immediately have a pydantic model created for it
   that is attached as an attribute.
"""

from typing import Any, Callable, Dict, List, Optional, Type
from pydantic import BaseModel, Field


class Function(BaseModel):
    """Schema for a function.

    :param name: Name of the function.
    :param description: Description/docstring of the function.
    :param parameters: Parameters of the function, including both args and kwargs.
    :param required: List of required parameter names.
    :param return_type: Return type annotation of the function.
    :param source_code: Source code of the function if available.
    """

    name: str = Field(..., description="Name of the function")
    description: str = Field(..., description="Description/docstring of the function")
    parameters: Dict[str, Any] = Field(
        default_factory=dict,
        description="Parameters of the function, including both args and kwargs",
    )
    required: List[str] = Field(
        default_factory=list, description="List of required parameter names"
    )
    return_type: Optional[Any] = Field(
        None, description="Return type annotation of the function"
    )
    source_code: Optional[str] = Field(
        None, description="Source code of the function if available"
    )
    _func: Optional[Callable] = None  # Private field to store the callable

    @classmethod
    def from_callable(cls, func: Callable) -> "Function":
        """Create a Function schema from a callable.

        :param func: The callable to create a schema from.
        :returns: A Function schema.
        """
        import inspect
        from typing import get_type_hints

        # Get function signature
        sig = inspect.signature(func)

        # Get docstring
        description = inspect.getdoc(func) or ""

        # Get parameters
        parameters = {}
        required = []
        for name, param in sig.parameters.items():
            if param.default == inspect.Parameter.empty:
                required.append(name)
            parameters[name] = (
                param.annotation if param.annotation != inspect.Parameter.empty else Any
            )

        # Get return type
        type_hints = get_type_hints(func)
        return_type = type_hints.get("return", None)

        # Get source code
        try:
            source_code = inspect.getsource(func)
        except (TypeError, OSError):
            source_code = None

        instance = cls(
            name=func.__name__,
            description=description,
            parameters=parameters,
            required=required,
            return_type=return_type,
            source_code=source_code,
        )
        instance._func = func  # Store the callable for later use
        return instance

    def to_pydantic_model(self) -> Type[BaseModel]:
        """Create a Pydantic model from the function schema.

        It needs to include the function name as well as all of the parameters
        to be passed into the function.

        :returns: A Pydantic model class representing the function parameters.
        """
        from pydantic import create_model, Field

        # Create field definitions for the model
        fields = {
            "function_name": (
                str,
                Field(default=self.name, description="Name of the function"),
            ),
        }

        # Add source code field if available
        if self.source_code is not None:
            fields["source_code"] = (
                str,
                Field(
                    default=self.source_code, description="Source code of the function"
                ),
            )

        # Add parameter fields
        for name, type_annotation in self.parameters.items():
            # If parameter is required, don't provide a default
            if name in self.required:
                fields[name] = (type_annotation, ...)
            else:
                fields[name] = (type_annotation, None)

        # Create and return the model class
        model_name = f"{self.name.title()}Parameters"
        return create_model(model_name, **fields)


def tool(func: Callable) -> Callable:
    """Decorator to create a tool from a function.

    :param func: The function to decorate.
    :returns: The decorated function with an attached Function schema.
    """
    # Create and attach the schema
    func.model = Function.from_callable(func).to_pydantic_model()
    func.json_schema = func.model.model_json_schema()
    return func


class FunctionToCall(BaseModel):
    """Pydantic model representing a function call with its parameters.

    :param function_name: Name of the function to call
    :param parameters: Dictionary mapping parameter names to their values
    """

    function_name: str
    parameters: dict[str, Any]


@tool
def add(a: int, b: int) -> int:
    """Add two integers."""
    return a + b



"""CLI for creating and keeping Markdown documentation up-to-date."""

try:
    import frontmatter
except ImportError:
    raise ImportError(
        "frontmatter is not installed. Please install it with `pip install llamabot[cli]`."
    )

from enum import Enum
from pathlib import Path

import requests
from pydantic import BaseModel, ConfigDict, Field, model_validator
from pyprojroot import here
from typer import Typer

import llamabot as lmb
from llamabot import prompt
from llamabot.bot.simplebot import SimpleBot
from llamabot.components.messages import AIMessage

app = Typer()


class DiataxisType(Enum):
    """The type of diataxis to use for the documentation.

    Based on the Diátaxis framework (https://diataxis.fr/), which defines four modes of documentation:
    - Tutorials: Learning-oriented, practical steps for newcomers
    - How-to guides: Problem-oriented, practical steps for specific goals
    - Reference: Information-oriented, theoretical knowledge
    - Explanation: Understanding-oriented, background and context
    """

    TUTORIAL = "tutorial"  # Learning-oriented guides for beginners
    HOWTO = "howto"  # Task-oriented guides for specific problems
    REFERENCE = "reference"  # Technical reference documentation
    EXPLANATION = "explanation"  # Background and conceptual guides


diataxis_sources = {
    DiataxisType.HOWTO: "https://raw.githubusercontent.com/evildmp/diataxis-documentation-framework/refs/heads/main/how-to-guides.rst",
    DiataxisType.REFERENCE: "https://raw.githubusercontent.com/evildmp/diataxis-documentation-framework/refs/heads/main/reference.rst",
    DiataxisType.TUTORIAL: "https://raw.githubusercontent.com/evildmp/diataxis-documentation-framework/refs/heads/main/tutorials.rst",
    DiataxisType.EXPLANATION: "https://raw.githubusercontent.com/evildmp/diataxis-documentation-framework/refs/heads/main/explanation.rst",
}


class MarkdownSourceFile(BaseModel):
    """Content related to a Markdown source file."""

    model_config = ConfigDict(arbitrary_types_allowed=True)
    file_path: Path = Field(..., description="Path to the Markdown source file.")
    post: frontmatter.Post = Field(..., description="The Markdown content.")
    diataxis_type: DiataxisType | None = Field(
        None, description="The type of diataxis to use for the documentation."
    )
    diataxis_source: str | None = Field(
        None, description="The source text for the diataxis documentation type."
    )

    linked_files: dict[str, str] = Field(
        {},
        description="Dictionary mapping linked files to their source texts. These must be relative to the repository root.",
    )
    raw_content: str = Field(
        "", description="The raw content of the Markdown source file."
    )

    def __init__(self, file_path: Path):
        # Initialize with basic file info first
        file_post = frontmatter.load(str(file_path))  # Convert Path to str
        super().__init__(file_path=file_path, post=file_post)

        # Handle linked files
        for fpath in self.post.get("linked_files", []):
            with open(here() / fpath, "r+") as f:
                self.linked_files[fpath] = f.read()

        # Handle diataxis type and source
        if self.post.get("diataxis_type"):
            self.diataxis_type = DiataxisType(self.post.get("diataxis_type"))
            if self.diataxis_type in diataxis_sources:
                response = requests.get(diataxis_sources[self.diataxis_type])
                if response.status_code == 200:
                    self.diataxis_source = response.text

        # Handle raw content
        with open(file_path, "r+") as f:
            self.raw_content = "".join(
                f"{i+1}: {line}" for i, line in enumerate(f.readlines())
            )

    def save(self):
        """Save the Markdown source file with the updated content."""

        with open(self.file_path, "w") as f:
            f.write(frontmatter.dumps(self.post))


class DocumentationContent(BaseModel):
    """Content related to documentation."""

    content: str = Field(..., description="The documentation content.")

    @model_validator(mode="after")
    def check_content(self):
        """Validate the content field."""
        if self.content.startswith("```") or self.content.endswith("```"):
            raise ValueError(
                "Documentation content should not start or end with triple backticks."
            )
        return self


class ModelValidatorWrapper(BaseModel):
    """Base class for model validators that wrap the model with additional fields."""

    status: bool = Field(..., description="Status indicator.")
    reasons: list[str] = Field(
        default_factory=list, description="Reasons for the status."
    )

    @model_validator(mode="after")
    def validate_status_and_reasons(self):
        """Validate the status and reasons fields.

        This function checks if the status and reasons fields are in a valid state.
        Validity is defined by:

        - status == True and reasons are not empty
        - status == False and reasons are empty
        """
        # The only valid states are status == True and reasons are not empty,
        # or that status == False and reasons are empty.
        if self.status and not self.reasons:
            raise ValueError("If status is True, reasons must be provided.")
        elif not self.status and self.reasons:
            raise ValueError("If status is False, reasons be empty.")
        return self


class DocumentationOutOfSyncWithSource(ModelValidatorWrapper):
    """Status indicating whether the documentation is out of sync with the source code."""

    status: bool = Field(
        ...,
        description="Whether the documentation is out of sync with the source code.",
    )
    reasons: list[str] = Field(
        default_factory=list,
        description="Reasons why the documentation is out of sync with the source code. Be specific.",
    )


class SourceContainsContentNotCoveredInDocs(ModelValidatorWrapper):
    """Status indicating whether the source contains content not covered in the documentation."""

    status: bool = Field(
        ...,
        description="Whether the source contains content not covered in the documentation.",
    )
    reasons: list[str] = Field(
        default_factory=list,
        description="Reasons why the source contains content not covered in the documentation. Be specific.",
    )


class DocsContainFactuallyIncorrectMaterial(ModelValidatorWrapper):
    """Status indicating whether docs contain factually incorrect material that contradicts the source code."""

    status: bool = Field(
        ...,
        description="Whether the documentation contains factually incorrect material that contradicts the source code.",
    )
    reasons: list[str] = Field(
        default_factory=list,
        description="The factually incorrect material that contradicts the source code. Specify the contradiction.",
    )


class DocsDoNotCoverIntendedMaterial(ModelValidatorWrapper):
    """Status indicating whether or not the documentation does not cover the intended material."""

    status: bool = Field(
        ...,
        description="Whether the intents of the documentation are not covered by the source. Return False if intents are adequately covered. Return True if some intents are not covered.",
    )
    reasons: list[str] = Field(
        default_factory=list,
        description="Reasons why the intents of the documentation are not covered by the source. Be specific, citing the intent that isn't covered.",
    )


class DocsOutOfDate(BaseModel):
    """Content related to documentation out of date."""

    source_not_covered: SourceContainsContentNotCoveredInDocs
    intents_not_covered: DocsDoNotCoverIntendedMaterial
    factually_inaccurate: DocsContainFactuallyIncorrectMaterial

    def __bool__(self):
        """Return True if any of the sub-models are True."""
        return (
            bool(self.source_not_covered)
            or bool(self.intents_not_covered)
            or bool(self.factually_inaccurate)
        )


@prompt(role="system")
def ood_checker_sysprompt() -> str:
    """You are an expert in documentation management.
    You will be provided information about a written documentation file,
    what the documentation is intended to convey,
    a list of source files that are related to the documentation,
    and their contents.
    """


@prompt(role="user")
def documentation_information(source_file: MarkdownSourceFile) -> str:
    """Here, I will provide you with contextual information to do your work.

    ## Referenced source files

    These are the source files to reference:

    {% for filename, content in source_file.linked_files.items() %}
    [[ {{ filename }} BEGINS ]]
    {{ content }}
    [[ {{ filename }} ENDS ]]
    {% endfor %}

    ## Documentation source file

    Here is the documentation in its current state, {{ source_file.file_path }}:

    [[ {{ source_file.file_path }} BEGINS ]]
    {% if source_file.post.content %}
    {{ source_file.post.content | safe }}
    {% else %}
    <The documentation is empty.>
    {% endif %}
    [[ {{ source_file.file_path }} ENDS ]]

    ## Intents about the documentation

    Here is the intent about the documentation:
    [[ INTENTS BEGINS ]]
    {% for intent in source_file.post.get("intents", []) %}- {{ intent }}
    {% endfor %}
    [[ INTENTS ENDS ]]

    {% if source_file.diataxis_type and source_file.diataxis_source %}
    ## Diataxis Framework Guide

    This documentation should follow the {{ source_file.diataxis_type.value }} style from the Diataxis framework.
    Here is the relevant guide:

    [[ DIATAXIS GUIDE BEGINS ]]
    {{ source_file.diataxis_source }}
    [[ DIATAXIS GUIDE ENDS ]]
    {% endif %}

    ## Instructions

    Now please write the docs as requested according to the intents and diataxis guide.
    Where relevant, leverage any commentary present in the source files.
    """


@prompt(role="user")
def ood_checker_information(ood_result: DocsOutOfDate) -> str:
    """Here is the out-of-date check result:

    [[ OOD CHECK RESULT BEGINS ]]
    {{ ood_result.model_dump_json() }}
    [[ OOD CHECK RESULT ENDS ]]
    """


@prompt(role="system")
def docwriter_sysprompt():
    """
    [[ Persona ]]
    You are an expert in documentation writing.

    [[ Context ]]
    You will be provided a documentation source,
    a list of what the documentation is intended to cover,
    and a list of linked source files.

    [[ Instructions ]]
    Based on the intended information that the documentation is supposed to cover
    and the content of linked source files,
    please edit (or create) the documentation,
    ensuring that the documentation matches the intents
    and has the correct content from the linked source files.
    """


def ood_checker_bot(model_name: str = "gpt-4o") -> lmb.StructuredBot:
    """Return a StructuredBot for the out-of-date checker."""
    return lmb.StructuredBot(
        system_prompt=ood_checker_sysprompt(),
        pydantic_model=DocsOutOfDate,
        model_name=model_name,
    )


def docwriter_bot(model_name: str = "gpt-4o") -> lmb.StructuredBot:
    """Return a StructuredBot for the documentation writer."""
    return lmb.StructuredBot(
        system_prompt=docwriter_sysprompt(),
        pydantic_model=DocumentationContent,
        model_name=model_name,
    )


@prompt(role="system")
def refine_bot_sysprompt():
    """
    You are an expert in documentation writing.
    You will be provided a documentation source that has been written
    using other sources as context.
    Those other sources are not going to be provided to you,
    only the documentation source that you are going to refine will be provided.
    Refine the documentation for clarity and logical flow.
    """


def refine_bot(model_name: str = "o1-preview") -> SimpleBot:
    """Return a SimpleBot for the documentation writer."""
    return lmb.SimpleBot(
        system_prompt=refine_bot_sysprompt(),
        model_name=model_name,
    )


@app.command()
def write(
    file_path: Path,
    from_scratch: bool = False,
    refine: bool = False,
    verbose: bool = False,
    ood_checker_model_name: str = "gpt-4o",
    docwriter_model_name: str = "gpt-4o",
    refiner_model_name: str = "o1-preview",
):
    """Write the documentation based on the given source file.

    The Markdown file should have frontmatter that looks like this:

    ```markdown
    ---
    intents:
    - Point 1 that the documentation should cover.
    - Point 2 that the documentation should cover.
    - ...
    linked_files:
    - path/to/relevant_file1.py
    - path/to/relevant_file2.toml
    - ...
    ---
    ```

    :param file_path: Path to the Markdown source file.
    :param from_scratch: Whether to start with a blank documentation.
    :param refine: Whether to refine the documentation.
    :param verbose: Whether to print the verbose output.
    :param ood_checker_model_name: The model name for the out-of-date checker.
    :param docwriter_model_name: The model name for the docwriter.
    :param refiner_model_name: The model name for the refiner.
    """
    src_file = MarkdownSourceFile(file_path)

    if from_scratch:
        src_file.post.content = ""

    ood_checker = ood_checker_bot(model_name=ood_checker_model_name)
    result: DocsOutOfDate = ood_checker(
        documentation_information(src_file), verbose=verbose
    )

    if (not src_file.post.content) or result:
        docwriter = docwriter_bot(model_name=docwriter_model_name)
        # print("-------")
        # print("OOD CHECKER INFORMATION:")
        # print(ood_checker_information(result))
        # print("-------")
        response: DocumentationContent = docwriter(
            lmb.user(
                documentation_information(src_file), ood_checker_information(result)
            ),
            verbose=verbose,
        )
        src_file.post.content = response.content

    if refine:
        refiner = refine_bot(model_name=refiner_model_name)
        response: AIMessage = refiner(src_file.post.content)
        src_file.post.content = response.content

    src_file.save()



"""Prompts for writing git stuff."""

from llamabot import SimpleBot
from llamabot.prompt_manager import prompt


def commitbot():
    """Return a commitbot instance.

    It is hard-coded to use mistral/mistral-medium as its model.
    This model is sufficient for the quality of commit messages,
    matching gpt-4-32k but being a fraction of the cost.

    :return: A commitbot instance.
    """
    return SimpleBot(
        "You are an expert user of Git.",
        model_name="gpt-4-0125-preview",
        stream_target="stdout",
    )


@prompt(role="user")
def write_commit_message(diff: str):
    """Please write a commit message for the following diff.

    {{ diff }}

    # noqa: DAR101

    Use the Conventional Commits specification to write the diff.

    [COMMIT MESSAGE BEGIN]
    <type>[optional scope]: <description>

    [optional body]

    [optional footer(s)]
    [COMMIT MESSAGE END]

    The commit contains the following structural elements,
    to communicate intent to the consumers of your library:

    fix: a commit of the type fix patches a bug in your codebase
        (this correlates with PATCH in Semantic Versioning).
    feat: a commit of the type feat introduces a new feature to the codebase
        (this correlates with MINOR in Semantic Versioning).
    BREAKING CHANGE: a commit that has a footer BREAKING CHANGE:,
        or appends a ! after the type/scope,
        introduces a breaking API change
        (correlating with MAJOR in Semantic Versioning).
        A BREAKING CHANGE can be part of commits of any type.

    types other than fix: and feat: are allowed,
    for example @commitlint/config-conventional
    (based on the Angular convention) recommends
    build:, chore:, ci:, docs:, style:, refactor:, perf:, test:, and others.

    footers other than BREAKING CHANGE: <description> may be provided
    and follow a convention similar to git trailer format.

    Additional types are not mandated by the Conventional Commits specification,
    and have no implicit effect in Semantic Versioning
    (unless they include a BREAKING CHANGE).
    A scope may be provided to a commit's type,
    to provide additional contextual information and is contained within parenthesis,
    e.g., feat(parser): add ability to parse arrays.
    Within the optional body section, prefer the use of bullet points.

    Final instructions:

    1. Do not fence the commit message with back-ticks or quotation marks.
    2. Do not add any other text except the commit message itself.
    3. Only write out the commit message.

    [BEGIN COMMIT MESSAGE]
    """


@prompt(role="user")
def compose_git_activity_report(log_info: str, hours: int) -> str:
    """Given the following git log information for the last {{ hours }} hours,
    please write a concise report summarizing the work done during this period.
    Highlight key changes, features, or fixes.
    Use markdown formatting.

    [[GIT LOG BEGIN]]
    {{ log_info }}
    [[GIT LOG END]]

    Please format your report as follows:

    1. Start with a brief summary of the overall work done in the specified time period.
    2. Use the following sections, omitting any that are not applicable:
       - New Features
       - Bug Fixes
       - Improvements
       - Documentation
       - Other Changes
    3. Under each section, use bullet points to list the relevant changes.
    4. For each bullet point, include:
       - A concise description of the change
       - The commit hash (first 7 characters)
       - The author's name (if available)
    5. If there are many similar changes, group them together and provide a summary.
    6. Highlight any significant or breaking changes.
    7. End with a brief conclusion or outlook, if appropriate.

    Example format:

    # Work Summary for the Last {{ hours }} Hours

    ## Overview
    [Brief summary of overall work]

    ## New Features
    - Implemented user authentication system (abc1234) (Jane Doe)
    - Added export functionality for reports (def5678) (John Smith)

    ## Bug Fixes
    - Fixed crash in data processing module (ghi9101) (Alice Johnson)

    ## Improvements
    - Optimized database queries for faster performance (jkl1121) (Bob Wilson)

    ## Conclusion
    [Brief conclusion or future outlook]
    """


@prompt(role="user")
def compose_release_notes(commit_log: str) -> str:
    """Here is a commit log:

    # noqa: DAR101

    {{ commit_log }}

    Please write for me the release notes.
    The notes should contain a human-readable summary
    of each new feature that was added.

    Follow the following format:

        ## Version <version number>

        <brief summary of the new version>

        ### New Features

        - <describe in plain English> (<commit's first 6 letters>) (<commit author>)
        - <describe in plain English> (<commit's first 6 letters>) (<commit author>)

        ### Bug Fixes

        - <describe in plain English> (<commit's first 6 letters>) (<commit author>)

        ### Deprecations

        - <describe in plain English> (<commit's first 6 letters>) (<commit author>)
    """



"""Prompt recorder class definition."""

import contextvars
import json
import hashlib
from datetime import datetime
from pathlib import Path
from typing import Any, Optional, List, Dict

from pyprojroot import here
from sqlalchemy import (
    Column,
    Connection,
    Engine,
    Float,
    ForeignKey,
    Integer,
    String,
    Text,
    create_engine,
    inspect,
    text,
)
from sqlalchemy.exc import OperationalError
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, relationship
from tenacity import retry, stop_after_attempt, wait_exponential

from llamabot.components.messages import BaseMessage
from llamabot.utils import find_or_set_db_path, get_object_name
from loguru import logger


prompt_recorder_var = contextvars.ContextVar("prompt_recorder")


class PromptRecorder:
    """Prompt recorder to support recording of prompts and responses."""

    def __init__(self):
        self.prompts: List[Dict[str, Any]] = []

    def __enter__(self):
        """Enter the context manager.

        :returns: The prompt recorder.
        """
        prompt_recorder_var.set(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit the context manager.

        :param exc_type: The exception type.
        :param exc_val: The exception value.
        :param exc_tb: The exception traceback.
        """
        prompt_recorder_var.set(None)

    def log(self, prompt: str, response: str):
        """Log the prompt and response in chat history.

        :param prompt: The human prompt.
        :param response: A the response from the bot.
        """
        self.prompts.append({"prompt": prompt, "response": response})

    def __repr__(self):
        """Return a string representation of the prompt recorder.

        :return: A string form of the prompts and responses as a dataframe.
        """
        import pandas as pd

        return pd.DataFrame(self.prompts).__str__()

    def _repr_html_(self):
        """Return an HTML representation of the prompt recorder.

        :return: We delegate to the _repr_html_ method of the pandas DataFrame class.
        """
        return self.dataframe()._repr_html_()

    def dataframe(self):
        """Return a pandas DataFrame representation of the prompt recorder.

        :return: A pandas DataFrame representation of the prompt recorder.
        """
        import pandas as pd

        return pd.DataFrame(self.prompts)

    def save(self, path: Path):
        """Save the prompt recorder to a path.

        :param path: The path to save the prompt recorder to.
        """
        path = Path(path)  # coerce to pathlib.Path
        with path.open("w+") as f:
            for prompt in self.prompts:
                f.write(f"**{prompt['prompt']}**\n\n{prompt['response']}\n\n")

    def panel(self):
        """Return a panel representation of the prompt recorder.

        :return: A panel representation of the prompt recorder.
        """
        import panel as pn

        global index
        index = 0
        pn.extension()

        next_button = pn.widgets.Button(name=">")
        prev_button = pn.widgets.Button(name="<")

        buttons = pn.Row(prev_button, next_button)

        prompt_header = pn.pane.Markdown("# Prompt")
        prompt_display = pn.pane.Markdown(self.prompts[index]["prompt"])

        prompt = pn.Column(prompt_header, prompt_display)

        response_header = pn.pane.Markdown("# Response")
        response_display = pn.pane.Markdown(self.prompts[index]["response"])
        response = pn.Column(response_header, response_display)

        display = pn.Row(prompt, response)

        def update_objects(index):
            """Update the prompt and response Markdown panes.

            :param index: The index of the prompt and response to update.
            """
            prompt_display.object = self.prompts[index]["prompt"]
            response_display.object = self.prompts[index]["response"]

        def next_button_callback(event):
            """Callback function for the next button.

            :param event: The click event.
            """
            global index
            index += 1
            if index > len(self.prompts) - 1:
                index = len(self.prompts) - 1

            update_objects(index)

        def prev_button_callback(event):
            """Callback function for the previous button.

            :param event: The click event.
            """
            global index
            index -= 1
            if index < 0:
                index = 0
            update_objects(index)

        next_button.on_click(next_button_callback)
        prev_button.on_click(prev_button_callback)

        return pn.Column(buttons, display)


def autorecord(prompt: str, response: str):
    """Record a prompt and response.

    This is intended to be called within every bot.
    If we are within a prompt recorder context,
    then the prompt recorder will record the prompt and response
    as specified in the function.

    :param prompt: The human prompt.
    :param response: A the response from the bot.
    """
    # Log the response.
    prompt_recorder: Optional[PromptRecorder] = prompt_recorder_var.get(None)
    if prompt_recorder:
        prompt_recorder.log(prompt, response)


# Design sqlite database for storing prompts and responses.
# Columns:
# - Python object name from globals()
# - Time stamp
# - Full message log as a JSON string of dictionaries.


Base = declarative_base()


class MessageLog(Base):
    """A log of a message exchange."""

    __tablename__ = "message_log"

    id = Column(Integer, primary_key=True, index=True)
    timestamp = Column(String, index=True)
    object_name = Column(String)
    model_name = Column(String)
    temperature = Column(Float)
    message_log = Column(Text)
    rating = Column(Integer, nullable=True)  # 1 for thumbs up, 0 for thumbs down

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._message_log_dict = None

    @property
    def message_log_dict(self) -> Dict[str, Any]:
        """
        Returns the message log as a dictionary.

        If the message log is not already parsed, it attempts to parse it from JSON.
        If parsing fails, it returns an empty dictionary.

        :return: The message log as a dictionary.
        """
        if not hasattr(self, "_message_log_dict"):
            try:
                self._message_log_dict = (
                    json.loads(self.message_log) if self.message_log else {}
                )
            except json.JSONDecodeError:
                self._message_log_dict = {}
        return self._message_log_dict


class Runs(Base):
    """A record of an experiment run."""

    __tablename__ = "runs"

    id = Column(Integer, primary_key=True)
    experiment_name = Column(String)
    timestamp = Column(String)
    run_metadata = Column(Text)
    run_data = Column(Text)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._run_data_dict = None

    @property
    def run_data_dict(self) -> Dict[str, Any]:
        """
        Returns the run data as a dictionary.

        If the run data is not already parsed, it attempts to parse it from JSON.
        If parsing fails, it returns an empty dictionary.

        :return: The run data as a dictionary.
        """
        if not hasattr(self, "_run_data_dict"):
            try:
                self._run_data_dict = json.loads(self.run_data) if self.run_data else {}
            except json.JSONDecodeError:
                self._run_data_dict = {}
        return self._run_data_dict


def upgrade_database(engine: Engine):
    """Upgrade the database schema."""
    # Create all tables from Base (now includes Runs)
    Base.metadata.create_all(engine)

    # Add new columns to existing tables if needed
    with engine.connect() as connection:
        inspector = inspect(engine)
        for table in [MessageLog, Prompt, Runs]:
            if inspector.has_table(table.__tablename__):
                existing_columns = [
                    c["name"] for c in inspector.get_columns(table.__tablename__)
                ]
                for column in table.__table__.columns:
                    if column.name not in existing_columns:
                        try:
                            add_column(connection, table.__tablename__, column)
                        except OperationalError as e:
                            print(f"Error adding column {column.name}: {e}")

    # Add rating column if it doesn't exist
    inspector = inspect(engine)
    if "message_log" in inspector.get_table_names():
        columns = [col["name"] for col in inspector.get_columns("message_log")]
        if "rating" not in columns:
            with engine.begin() as connection:
                connection.execute(
                    text("ALTER TABLE message_log ADD COLUMN rating INTEGER")
                )


def add_column(connection: Connection, table_name: str, column: Column):
    """
    Add a new column to an existing table.

    :param connection: SQLAlchemy connection
    :param table_name: Name of the table to modify
    :param column: Column object to add
    """
    column_name = column.name
    column_type = column.type.compile(connection.dialect)
    connection.execute(
        text(f"ALTER TABLE {table_name} ADD COLUMN {column_name} {column_type}")
    )


# Example usage:
# To add a new column in the future, you would do:
# add_column(engine, PromptResponseLog.__tablename__, Column('new_column_name', String))


def ensure_db_in_gitignore(db_path: Path) -> None:
    """Ensure the database file is listed in .gitignore.

    :param db_path: Path to the database file
    """
    try:
        repo_root = here()
        gitignore_path = repo_root / ".gitignore"
        db_filename = db_path.name

        if gitignore_path.exists():
            with open(gitignore_path, "r+") as f:
                content = f.read()
                if db_filename not in content:
                    f.write(f"\n# SQLite database\n{db_filename}\n")
        else:
            with open(gitignore_path, "w") as f:
                f.write(f"# SQLite database\n{db_filename}\n")
    except Exception as e:
        logger.debug(f"Could not update .gitignore: {e}")


@retry(stop=stop_after_attempt(5), wait=wait_exponential(multiplier=1, min=4, max=10))
def sqlite_log(
    obj: Any, messages: list[BaseMessage], db_path: Optional[Path] = None
) -> int:
    """Log messages to the sqlite database for further analysis.

    :param obj: The object to log the messages for.
    :param messages: The messages to log.
    :param db_path: The path to the database to use.
        If not specified, defaults to ~/.llamabot/message_log.db
    :return: ID of the created message log entry
    """
    # Set up the database path
    db_path = find_or_set_db_path(db_path)

    # Ensure database is in .gitignore if we're in a git repo
    ensure_db_in_gitignore(db_path)

    # Create the engine and session
    engine = create_engine(f"sqlite:///{db_path}")
    Base.metadata.create_all(engine)
    upgrade_database(engine)  # Add this line to ensure the database is upgraded
    Session = sessionmaker(bind=engine)
    session = Session()

    try:
        # Get the object name
        object_name = get_object_name(obj)

        # Get the current timestamp
        timestamp = datetime.now().isoformat()

        # Convert messages to a JSON string, including prompt_hash
        message_log = json.dumps(
            [
                {
                    "role": message.role,
                    "content": message.content,
                    "prompt_hash": (
                        message.prompt_hash if hasattr(message, "prompt_hash") else None
                    ),
                }
                for message in messages
            ]
        )

        # Create a new MessageLog instance
        new_log = MessageLog(
            object_name=object_name,
            timestamp=timestamp,
            message_log=message_log,
            model_name=obj.model_name,
            temperature=obj.temperature if hasattr(obj, "temperature") else None,
        )

        # Add the new log to the session and commit
        session.add(new_log)
        session.commit()

        # Get the ID before we do anything else
        log_id = new_log.id

        # If we're in an experiment context, add this message log to the run
        from .experiments import current_run

        experiment = current_run.get(None)
        if experiment is not None:
            experiment.add_message_log(log_id)

        return log_id

    finally:
        # Always close the session
        session.close()


class Prompt(Base):
    """A version-controlled prompt template."""

    __tablename__ = "prompts"

    id = Column(Integer, primary_key=True, index=True)
    hash = Column(String, unique=True, index=True)
    function_name = Column(String)
    template = Column(Text)
    previous_version_id = Column(Integer, ForeignKey("prompts.id"), nullable=True)
    previous_version = relationship(
        "Prompt", remote_side=[id], backref="next_version", uselist=False
    )


def hash_template(template: str) -> str:
    """Hash a prompt template."""
    return hashlib.sha256(template.encode()).hexdigest()


def store_prompt_version(
    session, template: str, function_name: str, previous_hash: Optional[str] = None
):
    """Store a new prompt version in the database."""
    logger.debug(f"Storing prompt version for function: {function_name}")
    template_hash = hash_template(template)

    existing_prompt = session.query(Prompt).filter_by(hash=template_hash).first()
    if existing_prompt:
        logger.debug(f"Existing prompt found with hash: {template_hash}")
        return existing_prompt

    previous_version = None
    if previous_hash:
        previous_version = session.query(Prompt).filter_by(hash=previous_hash).first()
        logger.debug(
            f"Previous version found: {previous_version.id if previous_version else None}"
        )

    new_prompt = Prompt(
        hash=template_hash,
        template=template,
        function_name=function_name,
        previous_version=previous_version,
    )
    session.add(new_prompt)
    logger.debug(f"New prompt added to session with hash: {template_hash}")
    session.flush()
    logger.debug(f"Session flushed, new prompt id: {new_prompt.id}")
    return new_prompt
